package dev.shadowsoffire.apotheosis.adventure.loot;

import dev.shadowsoffire.apotheosis.adventure.affix.Affix;
import java.util.Map;
import net.minecraft.world.item.ItemStack;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixRegistry;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixType;
import java.util.HashMap;
import net.minecraft.world.item.Items;
import com.google.common.base.Preconditions;
import dev.shadowsoffire.placebo.codec.CodecProvider;
import com.mojang.serialization.Codec;
import java.util.Iterator;
import java.util.Collection;
import com.google.common.collect.ImmutableList;
import java.util.function.Function;
import java.util.Comparator;
import dev.shadowsoffire.apotheosis.Apotheosis;
import net.minecraft.resources.ResourceLocation;
import com.google.common.collect.HashBiMap;
import java.util.ArrayList;
import dev.shadowsoffire.apotheosis.adventure.AdventureModule;
import net.minecraft.world.item.Item;
import com.google.common.collect.BiMap;
import dev.shadowsoffire.placebo.reload.DynamicHolder;
import java.util.List;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;

public class RarityRegistry extends WeightedDynamicRegistry<LootRarity>
{
    public static final RarityRegistry INSTANCE;
    protected List<DynamicHolder<LootRarity>> ordered;
    protected BiMap<Item, DynamicHolder<LootRarity>> materialMap;
    
    private RarityRegistry() {
        super(AdventureModule.LOGGER, "rarities", true, false);
        this.ordered = new ArrayList<DynamicHolder<LootRarity>>();
        this.materialMap = (BiMap<Item, DynamicHolder<LootRarity>>)HashBiMap.create();
    }
    
    public static boolean isMaterial(final Item item) {
        return getMaterialRarity(item).isBound();
    }
    
    public static DynamicHolder<LootRarity> getMaterialRarity(final Item item) {
        return (DynamicHolder<LootRarity>)RarityRegistry.INSTANCE.materialMap.getOrDefault((Object)item, RarityRegistry.INSTANCE.emptyHolder());
    }
    
    public static DynamicHolder<LootRarity> getMinRarity() {
        return RarityRegistry.INSTANCE.ordered.get(0);
    }
    
    public static DynamicHolder<LootRarity> getMaxRarity() {
        return RarityRegistry.INSTANCE.ordered.get(RarityRegistry.INSTANCE.ordered.size() - 1);
    }
    
    public static DynamicHolder<LootRarity> byOrdinal(final int i) {
        return RarityRegistry.INSTANCE.ordered.get(i);
    }
    
    public static ResourceLocation convertId(final String s) {
        return s.contains(":") ? new ResourceLocation(s) : Apotheosis.loc(s);
    }
    
    public static DynamicHolder<LootRarity> byLegacyId(final String s) {
        return (DynamicHolder<LootRarity>)RarityRegistry.INSTANCE.holder(convertId(s));
    }
    
    public static DynamicHolder<LootRarity> prev(final DynamicHolder<LootRarity> rarity) {
        if (rarity == getMinRarity()) {
            return rarity;
        }
        return byOrdinal(((LootRarity)rarity.get()).ordinal() - 1);
    }
    
    public static DynamicHolder<LootRarity> next(final DynamicHolder<LootRarity> rarity) {
        if (rarity == getMaxRarity()) {
            return rarity;
        }
        return byOrdinal(((LootRarity)rarity.get()).ordinal() + 1);
    }
    
    public List<DynamicHolder<LootRarity>> getOrderedRarities() {
        return this.ordered;
    }
    
    protected void beginReload() {
        super.beginReload();
        this.ordered = new ArrayList<DynamicHolder<LootRarity>>();
        this.materialMap = (BiMap<Item, DynamicHolder<LootRarity>>)HashBiMap.create();
    }
    
    protected void onReload() {
        super.onReload();
        this.ordered = this.registry.values().stream().sorted(Comparator.comparing((Function<? super T, ? extends Comparable>)LootRarity::ordinal)).map(this::holder).toList();
        int lastOrdinal = -1;
        for (DynamicHolder<LootRarity> r : this.ordered) {
            if (((LootRarity)r.get()).ordinal() != lastOrdinal + 1) {
                final StringBuilder sb = new StringBuilder();
                sb.append("Rarity ordinal order is inconsistent. The ordinals must start at zero and be continuous up to the max value.\n");
                for (DynamicHolder<LootRarity> rarity : this.ordered) {
                    sb.append(rarity.getId() + " | " + ((LootRarity)rarity.get()).ordinal());
                }
                throw new RuntimeException(sb.toString());
            }
            lastOrdinal = ((LootRarity)r.get()).ordinal();
        }
        for (final DynamicHolder<LootRarity> r : this.ordered) {
            final DynamicHolder<LootRarity> old = this.materialMap.put(((LootRarity)r.get()).getMaterial(), r);
            if (old != null) {
                throw new RuntimeException("Two rarities may not share the same rarity material: " + r.getId() + " conflicts with " + old.getId());
            }
        }
        this.ordered = (List<DynamicHolder<LootRarity>>)ImmutableList.copyOf((Collection<?>)this.ordered);
    }
    
    protected void registerBuiltinCodecs() {
        this.registerDefaultCodec(Apotheosis.loc("rarity"), (Codec)LootRarity.LOAD_CODEC);
    }
    
    protected void validateItem(final ResourceLocation key, final LootRarity item) {
        super.validateItem(key, (CodecProvider)item);
        Preconditions.checkNotNull(item.getColor());
        Preconditions.checkArgument(item.getMaterial() != null && item.getMaterial() != Items.f_41852_);
        Preconditions.checkArgument(item.getWeight() >= 0, (Object)"A rarity may not have negative weight!");
        Preconditions.checkArgument(item.getQuality() >= 0.0f, (Object)"A rarity may not have negative quality!");
        Preconditions.checkArgument(!item.getRules().isEmpty(), (Object)"A rarity may not have no rules!");
    }
    
    public void validateLootRules() {
        for (LootRarity rarity : this.registry.values()) {
            final Map<AffixType, List<LootRarity.LootRule>> sorted = new HashMap<AffixType, List<LootRarity.LootRule>>();
            rarity.getRules().stream().filter(r -> r.type().needsValidation()).forEach(rule -> {
                sorted.computeIfAbsent(rule.type(), r -> new ArrayList());
                sorted.get(rule.type()).add(rule);
                return;
            });
            sorted.forEach((type, rules) -> {
                LootCategory.VALUES.iterator();
                final Iterator iterator2;
                while (iterator2.hasNext()) {
                    final LootCategory cat = iterator2.next();
                    if (cat.isNone()) {
                        continue;
                    }
                    else {
                        final List<Affix> affixes = AffixRegistry.INSTANCE.getValues().stream().filter(a -> a.canApplyTo(ItemStack.f_41583_, cat, rarity) && a.getType() == type).toList();
                        if (affixes.size() < rules.size()) {
                            final StringBuilder errMsg = new StringBuilder();
                            errMsg.append("Insufficient number of affixes to satisfy the loot rules (ignoring backup rules) of rarity " + this.getKey((CodecProvider)rarity) + " for category " + cat.getName());
                            errMsg.append("Required: " + rules.size());
                            errMsg.append("; Provided: " + affixes.size());
                            AdventureModule.LOGGER.warn(errMsg.toString());
                        }
                        else {
                            continue;
                        }
                    }
                }
            });
        }
    }
    
    static {
        INSTANCE = new RarityRegistry();
    }
}
